   --------------------------------------------------------------------
--|  Package  : Abstract Binary Tree                                   |
   --------------------------------------------------------------------
--|  Abstract :                                                        |
--|    Provides a simple (non keyed) binary tree with generic objects. |
   --------------------------------------------------------------------
--|  File            : bintree.ads             Version : 1.0           |
--|  Compiler/System : Meridian Adavantage IBM PS/2                    |
--|  Author          : John Dalbey             Date : 4/7/90           |
--|  References      : Stubbs & Webre                                  |
   --------------------------------------------------------------------
--|  Limitations     : Insert can't deal with parents.                 |
--|                     (See notes below)                              |
   --------------------------------------------------------------------

--  ELEMENTS: The elements of the binary tree are generic objects.
--
--  STRUCTURE: A binary tree is either empty or is a node, called the
--  root node, together with two binary trees.  These two binary trees
--  are disjoint from each other and from the root node and are called
--  the left and right subtrees of the root node.
--  Note that a binary tree has a hierarchical structure.  There is
--  (unless the tree is empty) a unique root node.  Each node except
--  the root has a unique parent.  Each node has zero, one, or two
--  children.  A child is designated as either the left or right
--  child of its parent.
--
--  DOMAIN:  The number of nodes in a binary tree is bounded.

GENERIC
   TYPE Object IS PRIVATE;              -- The objects in the tree

PACKAGE Binary_tree IS

   TYPE BinaryTree IS PRIVATE;          -- The binary tree type
   TYPE Relative IS (leftChild, rightChild, root, parent);


--  OPERATIONS: Operations are provided for inserting new nodes
--       retrieving the value of a node, and for navigating around
--       the tree.  If the binary tree is not empty, then one
--       node is designated the CURRENT node.

PROCEDURE Insert (A_Tree: IN OUT BinaryTree;
                  Item  : IN     Object;
                  Where : IN     Relative;
                  Result:    OUT BOOLEAN);
-- PURPOSE: Insert a new node into the tree.
-- ASSUMES: Item has a value to be inserted.
--          Where has the position relative to CURRENT of the desired new node.
--          IF A_Tree is empty then Where = root
--          OTHERWISE A_Tree is not empty and Where /= root
-- RESULTS: IF Where = parent, Result is False (this operation not allowed)
--          Otherwise, Result is True, and Item has been inserted.
--          CURRENT becomes the new node.
-- NOTE: If there previously existed a child at WHERE, it is lost!.




PROCEDURE Retrieve (A_Tree: IN     BinaryTree;
                    Item  :    OUT Object);
-- ASSUMES: A_Tree is not empty.
-- RESULTS: Item is the value of the element in the CURRENT node.



PROCEDURE Find (A_Tree: IN OUT BinaryTree;
                Where : IN     Relative;
                Result:    OUT BOOLEAN);
-- ASSUMES: Where has a value of the relative to find.
--          A_Tree is not empty.
-- RESULTS: CURRENT is "moved" to the node determined by Where.
--          If the move is successful, TRUE is returned.
--          If the move is not successful (because there wasn't a
--          left or right child and you tried to move that way,
--          FALSE is returned.


FUNCTION IsLeaf (A_Tree: IN BinaryTree) RETURN BOOLEAN;
-- ASSUMES: nothing
-- RESULTS: True if CURRENT is a leaf node, FALSE otherwise
--          False if A_Tree is empty.


FUNCTION IsEmpty (A_Tree: IN BinaryTree) RETURN BOOLEAN;
-- ASSUMES: nothing
-- RESULTS: Returns TRUE is A_Tree contains zero nodes, else returns FALSE.


PROCEDURE MakeEmpty (A_Tree: IN OUT BinaryTree);
-- ASSUMES: nothing
-- RESULTS: A_Tree contains zero nodes.

PRIVATE
   TYPE node;                            -- incomplete declaration required
   TYPE node_ptr is access node;         -- pointer declaration
   TYPE node is                          -- full declaration of node type
         record
            the_object: Object;     -- data in the node
            left  : node_ptr;       -- pointer to left child
            right : node_ptr;       -- pointer to right child
         end record;

   TYPE BinaryTree is                     -- full declaration of tree type
         record
            treeRoot: node_ptr;     -- root of the tree
            Current : node_ptr;     -- the CURRENT node
         end record;

END Binary_Tree;
